
#include <stdio.h>
#include <assert.h>
#include <time.h>
#include <math.h>

#include "TLibCommon/TComRom.h"
#include "TLibCommon/TComARCFilter.h"

using namespace std;

//! \ingroup TLibEncoder
//! \{

#define LUMA    (0)
#define CHROMA  (1)

#define Y   (0)
#define U   (1)
#define V   (2)

#define MARGIN_X  (20)
#define MARGIN_Y  (20)

#if ARC_FILTERS
  ARCfilter *g_arcFilter = NULL;
#endif

//======================================================
// Up/Down-scaling Filters
//======================================================
//----------------------
//downScale2x2 
//----------------------
#define NTAPS_DOWN_SCALE    (12)
Short g_downScaleFilter[NTAPS_DOWN_SCALE] =
{
  4, 0, -12, 0, 40, 64, 40, 0, -12, 0, 4, 0       // sum = 128
};
Int g_downScaleFilter_ifFilterPrec=7;

Short g_downScaleFilterHalfPhase[NTAPS_DOWN_SCALE] =
{
  1, 3, -7, -15, 33, 113, 113, 33, -15, -7, 3, 1  // sum = 256
};
Int g_downScaleFilterHalfPhase_ifFilterPrec=8;

//----------------------
//downScale2x2 
//----------------------
#define NTAPS_UP_SCALE    (4)
Short g_upScaleFilter[2][NTAPS_UP_SCALE] =
{
  { 0, 64,  0,  0}, // phase 0
  {-8, 40, 40, -8}, // phase 1
};
Int g_upScaleFilter_ifFilterPrec=6;

#define NTAPS_UP_SCALE_QUARTER_PHASE    (8)
Short g_upScaleFilterQuarterPhase[2][NTAPS_UP_SCALE_QUARTER_PHASE] =
{
  { 1,    -6,   25,  130, -33,  11, 0, 0 },       // phase 0 - sum = 128
  { 0, 11,  -33,  130,   25,  -6,   1, 0     }  // phase 1 - sum = 128
};
Int g_upScaleFilterQuarterPhase_ifFilterPrec=7;

//==================================================================
//  DOWN-SCALING(downScale2x2) 2x2 + MV Interpolation Filters
//==================================================================
#define NTAPS_CHROMA_DOWN_SCALE_MV_INTER  (19)
#define NTAPS_LUMA_DOWN_SCALE_MV_INTER    (27)

//==========================================================================
//  DOWN-SCALING(downScale2x2HalfPhase) 2x2 + MV Interpolation Filters
//==========================================================================
Short g_Luma_downScale2x2_mvInter_filter[4][NTAPS_LUMA_DOWN_SCALE_MV_INTER] = {

    {1, 3, -7, -15, 33, 113, 113, 33, -15, -7, 3, 1, 0,   0,0,0,0,0, 0,0,0,0,0, 0,0,0,0  },  // frac=0	(sum=256)

    {0, -1, -3, 11, 27, -71, -203, 147, 743, -252, -1810, 597, 5925, 7313, 3987, 849, -435, -613, -113, 239, 85, -30, -12, 3, 1, 0, 0  },  // frac=1	(sum=16384)

    {0, -1, -3, 11, 27, -72, -206, 136, 704, -136, -1584, -277, 3495, 6098, 6098, 3495, -277, -1584, -136, 704, 136, -206, -72, 27, 11, -3, -1,  },  // frac=2	(sum=16384)

    {0, 0, 0, 1, 3, -12, -30, 85, 239, -113, -613, -435, 849, 3987, 7313, 5925, 597, -1810, -252, 743, 147, -203, -71, 27, 11, -3, -1  }   // frac=3	(sum=16384)

};

Int g_Luma_downScale2x2_mvInter_nbCoeff[4] = 
{
	13, 27, 27, 27
};

Int g_Luma_downScale2x2_mvInter_ifFilterPrec[4] = 
{
	8, 14, 14, 14  
};

Short g_Chroma_downScale2x2_mvInter_filter[8][NTAPS_CHROMA_DOWN_SCALE_MV_INTER] = {

    {1, 3, -7, -15, 33, 113, 113, 33, -15, -7, 3, 1, 0  },  // frac=0	(sum=256)

    {0, -1, -3, 36, 102, -231, -533, 808, 3166, 3464, 1544, 94, -152, -101, -39, 30, 12, -3, -1  },  // frac=1	(sum=8192)

    {0, -2, -6, 41, 111, -247, -607, 608, 2862, 3352, 1824, 460, -40, -152, -62, 39, 15, -3, -1  },  // frac=2	(sum=8192)

    {0, -3, -9, 44, 114, -246, -642, 320, 2284, 3120, 2392, 1162, 72, -367, -141, 72, 28, -6, -2  },  // frac=3	(sum=8192)

    {0, -1, -3, 16, 42, -87, -221, 120, 846, 1336, 1336, 846, 120, -221, -87, 42, 16, -3, -1  },  // frac=4	(sum=4096)

    {0, -2, -6, 28, 72, -141, -367, 72, 1162, 2392, 3120, 2284, 320, -642, -246, 114, 44, -9, -3  },  // frac=5	(sum=8192)

    {0, -1, -3, 15, 39, -62, -152, -40, 460, 1824, 3352, 2862, 608, -607, -247, 111, 41, -6, -2  },  // frac=6	(sum=8192)

    {0, -1, -3, 12, 30, -39, -101, -152, 94, 1544, 3464, 3166, 808, -533, -231, 102, 36, -3, -1  }   // frac=7	(sum=8192)

};

Int g_Chroma_downScale2x2_mvInter_nbCoeff[8] = 
{
	13, 19, 19, 19, 19, 19, 19, 19
};

Int g_Chroma_downScale2x2_mvInter_ifFilterPrec[8] = 
{
	8, 13, 13, 13, 12, 13, 13, 13  
};


#define NTAPS_LUMA_UP_SCALE_MV_INTER    (13)
#define NTAPS_CHROMA_UP_SCALE_MV_INTER  (11)
//===================================================
//  UP-SCALING(QuarterPhase) 2x2 + MV Interpolation Filters
//===================================================
Short g_Luma_upScale2x2_mvInter_filter[2][4][NTAPS_LUMA_UP_SCALE_MV_INTER] = {
  { // phase 0 :
    { 1, -6, 25, 130, -33, 11, 0, 0,     0,0,0,0,0 },  // frac=0	(sum=128)

    {0, -7, -43, 134, 125, 9216, -1906, 716, -44, 1, 0, 0, 0  },  // frac=1	(sum=8192)

    {0, -7, -72, 522, -1146, 9598, -1146, 522, -72, -7, 0, 0, 0  },  // frac=2	(sum=8192)

    {0, 1, -44, 716, -1906, 9216, 125, 134, -43, -7, 0, 0, 0  }   // frac=3	(sum=8192)

  },
  { // phase 1 :
    {0, 11, -33, 130, 25, -6, 1, 0,      0,0,0,0,0  },  // frac=0	(sum=128)

    {0, -1, 40, 558, -1930, 6957, 3330, -1010, 242, 6, 0, 0, 0  },  // frac=1	(sum=8192)

    {0, -1, 39, 389, -1562, 5231, 5231, -1562, 389, 39, -1, 0, 0  },  // frac=2	(sum=8192)

    {0, 0, 6, 242, -1010, 3330, 6957, -1930, 558, 40, -1, 0, 0  }   // frac=3	(sum=8192)

  } 
};

Int g_Luma_upScale2x2_mvInter_nbCoeff[2][4] = 
{
  { 8, 12, 12, 12  },
  { 8, 12, 12, 12  } 
};

Int g_Luma_upScale2x2_mvInter_ifFilterPrec[2][4] = 
{
  { 7, 13, 13, 13  },
  { 7, 13, 13, 13  } 
};

Short g_Chroma_upScale2x2_mvInter_filter[2][8][NTAPS_CHROMA_UP_SCALE_MV_INTER] = {
  { // phase 0 :
    {1, -6, 25, 130, -33, 11, 0, 0,   0,0,0   },  // frac=0	(sum=128)

    {0, 18, -87, 436, 4370, -956, 321, -6, 0, 0, 0  },  // frac=1	(sum=4096)

    {0, 5, -9, 157, 4475, -809, 280, -3, 0, 0, 0  },  // frac=2	(sum=4096)

    {0, -10, 113, -265, 4685, -651, 232, -8, 0, 0, 0  },  // frac=3	(sum=4096)

    {0, -2, 77, -196, 2290, -196, 77, -2, 0, 0, 0  },  // frac=4	(sum=2048)

    {0, -8, 232, -651, 4685, -265, 113, -10, 0, 0, 0  },  // frac=5	(sum=4096)

    {0, -3, 280, -809, 4475, 157, -9, 5, 0, 0, 0  },  // frac=6	(sum=4096)

    {0, -6, 321, -956, 4370, 436, -87, 18, 0, 0, 0  }   // frac=7	(sum=4096)

  },
  { // phase 1 :
    {0, 11, -33, 130, 25, -6, 1, 0,   0,0,0  },  // frac=0	(sum=128)

    {0, -1, 330, -1023, 3798, 1278, -375, 90, -1, 0, 0  },  // frac=1	(sum=4096)

    {0, -2, 317, -1000, 3483, 1651, -473, 121, -1, 0, 0  },  // frac=2	(sum=4096)

    {0, -3, 285, -940, 3016, 2234, -683, 189, -2, 0, 0  },  // frac=3	(sum=4096)

    {0, -1, 114, -387, 1298, 1298, -387, 114, -1, 0, 0  },  // frac=4	(sum=2048)

    {0, -2, 189, -683, 2234, 3016, -940, 285, -3, 0, 0  },  // frac=5	(sum=4096)

    {0, -1, 121, -473, 1651, 3483, -1000, 317, -2, 0, 0  },  // frac=6	(sum=4096)

    {0, -1, 90, -375, 1278, 3798, -1023, 330, -1, 0, 0  }   // frac=7	(sum=4096)

  } 
};

Int g_Chroma_upScale2x2_mvInter_nbCoeff[2][8] = 
{
  { 8, 10, 10, 10, 10, 10, 10, 10  },
  { 8, 10, 10, 10, 10, 10, 10, 10  } 
};

Int g_Chroma_upScale2x2_mvInter_ifFilterPrec[2][8] = 
{
  { 7, 12, 12, 12, 11, 12, 12, 12  },
  { 7, 12, 12, 12, 11, 12, 12, 12  } 
};

// ====================================================================================================================
// Constructor / destructor / initialization / destroy
// ====================================================================================================================

ARCfilter::ARCfilter(Int width, Int height)
{
  m_sizeX   = width;
  m_sizeY   = height;
  m_marginX = MARGIN_X;
  m_marginY = MARGIN_X;

  m_stride  = m_sizeX + 2*m_marginX;
  m_offset  = m_marginY*m_stride + m_marginX;

  m_pTemp = new Pel[ (m_sizeX+2*m_marginX) * (m_sizeY+2*m_marginY) ];

  m_downScaleFilter               = NULL;
  m_downScaleFilter_ifFilterPrec  = 0;

  m_upScaleFilter[0]              = NULL;
  m_upScaleFilter[1]              = NULL;
  m_upScaleFilter_ifFilterPrec    = 0;
  m_upScaleFilter_nTaps           = 0;
}

ARCfilter::~ARCfilter()
{
  delete m_pTemp;
}

Void  ARCfilter::razTemp()
{
  memset(m_pTemp, 0, sizeof(Pel) * (m_sizeX+2*m_marginX) * (m_sizeY+2*m_marginY));
}

Void  ARCfilter::razTemp(Int Sx, Int Sy)
{
  for ( Int i=0 ; i<(Sy+2*m_marginY) ; i++ )
    memset(m_pTemp + i*m_stride, 0, sizeof(Pel) * (Sx+2*m_marginX) );
}

Void  ARCfilter::setSize(Int width, Int height)
{
  if ( m_sizeX<width || m_sizeY<height ) {
    if ( m_pTemp )  delete m_pTemp;

    m_sizeX = width;
    m_sizeY = height;
    m_stride  = m_sizeX + 2*m_marginX;
    m_offset  = m_marginY*m_stride + m_marginX;

    m_pTemp = new Pel[ (m_sizeX+2*m_marginX) * (m_sizeY+2*m_marginY) ];
  }
}

Void  ARCfilter::freeTmpBuffer()
{
  delete m_pTemp; m_pTemp = NULL;

  m_sizeX = 0;
  m_sizeY = 0;
}

Int getSize(Int size, Int mode, Int isChroma)
{
  size = (mode==DOWN) ? (size>>1) : (size<<1);
  size = (isChroma)   ? (size>>1) : (size);
  return(size);
}

Int ARCfilter::getMode(Int iCurSizeIdx, Int iRefSizeIdx)
{
  Int mode;

  if ( iCurSizeIdx==0 && iRefSizeIdx==0 )       mode = SAME;
  else  if ( iCurSizeIdx==1 && iRefSizeIdx==0 ) mode = DOWN;
  else  if ( iCurSizeIdx==0 && iRefSizeIdx==1 ) mode = UP;
  else {
    printf("ARCfilter::setMode(iCurSizeIdx=%d, iRefSizeIdx=%d) => Not supported.\n", iCurSizeIdx, iRefSizeIdx);
    exit(0);
  }

  return( mode );
}

//  width, height:  size of the output
Void  ARCfilter::filter_ARC(Short const *src, Int srcStride, Short *dst, Int dstStride, Int width, Int height, StructFilterARC *stFilter)
{
  Int row, col, scol;
  
  Short *coeff        = stFilter->coeff[0];
  Int   N             = stFilter->N[0];
  Int   nbPhase       = stFilter->nbPhase;
  Bool  isVertical    = stFilter->isVertical;
  Bool  isLast        = stFilter->isLast;
  Bool  isFirst       = stFilter->isFirst;
  Int   cs            = stFilter->cs;
  Int   rs            = stFilter->rs;
  Int   ifFilterPrec  = stFilter->ifFilterPrec;

  assert( ifFilterPrec >= IF_FILTER_PREC );

  Short *c = coeff;

  Int cStride = ( isVertical ) ? (srcStride) : 1;
  src -= ( N/2 - 1 ) * cStride;

  Int offset;
  Short maxVal;
  Int headRoom = IF_INTERNAL_PREC - (g_uiBitDepth + g_uiBitIncrement);
  Int shift = ifFilterPrec;
  if ( isLast )
  {
    shift += (isFirst) ? 0 : headRoom;
    offset = 1 << (shift - 1);
    offset += (isFirst) ? 0 : IF_INTERNAL_OFFS << ifFilterPrec;
    maxVal = g_uiIBDI_MAX;
  }
  else
  {
    shift -= (isFirst) ? headRoom : 0;
    offset = (isFirst) ? -IF_INTERNAL_OFFS << shift : 0;
    maxVal = 0;
  }
  
  for (row = 0; row < height; row++)
  {
    for (col = 0, scol = 0; col < width; col++, scol+=cs)
    {
      Int sum=0;
      
      for ( int n=0 ; n<N ; n++ )
        sum += src[ scol + n * cStride] * c[n];
      
      Short val = ( sum + offset ) >> shift;

      if ( isLast )
      {
        val = ( val < 0 ) ? 0 : val;
        val = ( val > maxVal ) ? maxVal : val;        
      }
      dst[col] = val;
    }
    
    src += srcStride * rs;
    dst += dstStride;
  }    
}

Void  ARCfilter::filter_ARC_upScale_phasis(Short const *src, Int srcStride, Short *dst, Int dstStride, Int width, Int height, StructFilterARC *stFilter, Int phasis1st)
{
  Int srow, col, scol, phase;
  
  assert( stFilter->N[0] == stFilter->N[1] );

  Short *coeff        = NULL;
  Int   N             = stFilter->N[phasis1st];
  Int   nbPhase       = stFilter->nbPhase;
  Bool  isVertical    = stFilter->isVertical;
  Bool  isLast        = stFilter->isLast;
  Bool  isFirst       = stFilter->isFirst;
  Int   cs            = stFilter->cs;
  Int   rs            = stFilter->rs;
  Int   ifFilterPrec  = stFilter->ifFilterPrec;

  assert( ifFilterPrec >= IF_FILTER_PREC );

  Short *c = NULL;

  Int cStride = ( isVertical ) ? (srcStride) : 1;
  Int dStride = ( isVertical ) ? (dstStride) : 1;
  src -= ( N/2 - 1 ) * cStride;

  Int sheight = ( isVertical ) ? (height / nbPhase) : (height);
  Int swidth  = ( isVertical ) ? (width) : (width / nbPhase);

  Int offset;
  Short maxVal;
  Int headRoom = IF_INTERNAL_PREC - (g_uiBitDepth + g_uiBitIncrement);
  Int shift = ifFilterPrec;
  if ( isLast )
  {
    shift += (isFirst) ? 0 : headRoom;
    offset = 1 << (shift - 1);
    offset += (isFirst) ? 0 : IF_INTERNAL_OFFS << ifFilterPrec;
    maxVal = g_uiIBDI_MAX;
  }
  else
  {
    shift -= (isFirst) ? headRoom : 0;
    offset = (isFirst) ? -IF_INTERNAL_OFFS << shift : 0;
    maxVal = 0;
  }
  
  int np[2]={0,0};
  if ( phasis1st )
    np[1] = 1;

  for (srow = 0; srow < sheight; srow++)
  {
    for (scol = 0, phase=phasis1st ; scol < swidth ; scol++ )
    {

      for ( int iphase=0 ; iphase<nbPhase ; iphase++ ) {

        c = stFilter->coeff[phase];
        N = stFilter->N[phase];

        Int sum=0;
        
        for ( int n=0 ; n<N ; n++ )
          sum += src[ scol + (n+np[iphase]) * cStride] * c[n];
        
        Short val = ( sum + offset ) >> shift;

        if ( isLast )
        {
          val = ( val < 0 ) ? 0 : val;
          val = ( val > maxVal ) ? maxVal : val;        
        }

        col = ( isVertical ) ? (scol + iphase*dstStride) : (nbPhase*scol + iphase);

        dst[col] = val;

        phase = (phase<(nbPhase-1)) ? (phase+1) : (0);  // next phase

      } // for iPhase

    } // for col

    src += srcStride * rs;
    dst += ( isVertical ) ? (nbPhase*dstStride) : (dstStride);
  } // for row

}

Void  ARCfilter::extendBorderUpDown(Pel *src, Int stride, Int width, Int height, Int marginY)
{
  // Up Border :
  for ( int i=1 ; i<marginY ; i++ ) {
    memcpy(src - i*stride, src, sizeof(Pel)*width);
  }
  src += (height-1)*stride;
  // Down Border :
  for ( int i=1 ; i<marginY ; i++ ) {
    memcpy(src + i*stride, src, sizeof(Pel)*width);
  }
}

Void  ARCfilter::extendBorderLeftRight(Pel *src, Int stride, Int width, Int height, Int marginX)
{
  Pel *p=src;

  for ( int i=0 ; i<height ; i++ ) {
    std::fill(p - marginX, p, p[0] );                       // Left border
    std::fill(p + width, p + width + marginX, p[width-1] ); // Right border
    p+= stride;
  }
}

Void  ARCfilter::setFilter( Int mode, Bool isFirst, Bool isLast, Bool isVertical, StructFilterARC &stFilter)
{
  stFilter.nbPhase      = (mode == DOWN) ? (1) : (2);
  stFilter.coeff[0]     = (mode == DOWN) ? m_downScaleFilter       : m_upScaleFilter[0];
  stFilter.coeff[1]     = (mode == DOWN) ? NULL                    : m_upScaleFilter[1];
  stFilter.N[0]         = (mode == DOWN) ? m_downScaleFilter_nTaps : m_upScaleFilter_nTaps;
  stFilter.N[1]         = (mode == DOWN) ? (0)                     : m_upScaleFilter_nTaps;
  stFilter.isFirst      = isFirst;
  stFilter.isLast       = isLast;
  if ( isVertical ) {
    stFilter.cs           = 1;
    stFilter.rs           = (mode == DOWN) ? (2) : (1);
  }
  else {
    stFilter.cs           = (mode == DOWN) ? (2) : (1);
    stFilter.rs           = 1;
  }
  stFilter.ifFilterPrec = (mode == DOWN) ? (m_downScaleFilter_ifFilterPrec) : (m_upScaleFilter_ifFilterPrec);
  stFilter.isVertical   = isVertical;
}

Void  ARCfilter::scalePic( TComPicYuv* pcYuvSrc, TComPicYuv* pcYuvDst , Int mode, Bool isRef)
{
  m_downScaleFilter               = (isRef) ? (g_downScaleFilterHalfPhase)              : (g_downScaleFilter);
  m_downScaleFilter_ifFilterPrec  = (isRef) ? (g_downScaleFilterHalfPhase_ifFilterPrec) : (g_downScaleFilter_ifFilterPrec);
  m_downScaleFilter_nTaps         = (isRef) ? (NTAPS_DOWN_SCALE)                        : (NTAPS_DOWN_SCALE);

  m_upScaleFilter[0]              = (isRef) ? (g_upScaleFilterQuarterPhase[0])          : (g_upScaleFilter[0]);
  m_upScaleFilter[1]              = (isRef) ? (g_upScaleFilterQuarterPhase[1])          : (g_upScaleFilter[1]);
  m_upScaleFilter_ifFilterPrec    = (isRef) ? (g_upScaleFilterQuarterPhase_ifFilterPrec): (g_upScaleFilter_ifFilterPrec);
  m_upScaleFilter_nTaps           = (isRef) ? (NTAPS_UP_SCALE_QUARTER_PHASE)            : (NTAPS_UP_SCALE);

  setFilter( mode, true, false, true,   m_stFilterV);
  setFilter( mode, false, true, false,  m_stFilterH);

  pcYuvSrc->setBorderExtension(false);
  pcYuvSrc->extendPicBorder();

  filter( pcYuvSrc, pcYuvDst , mode, isRef);

  pcYuvDst->setBorderExtension(false);
  pcYuvDst->extendPicBorder();
}

Void  ARCfilter::filter( TComPicYuv* pcYuvSrc, TComPicYuv* pcYuvDst , Int mode, Bool isRef)
{
  int   srcSizeX, srcSizeY, srcStride;
  int   dstSizeX, dstSizeY, dstStride;
  Short *src, *dst;
  Bool  bChroma = true;
  Short *upScaleFilter[2] = {NULL,NULL};

  pcYuvSrc->setBorderExtension(false);
  pcYuvSrc->extendPicBorder();

  if ( mode == DOWN ) setSize(pcYuvSrc->getWidth(),pcYuvSrc->getHeight());
  else                setSize(pcYuvDst->getWidth(),pcYuvDst->getHeight());

  razTemp();

  //--------------------------------------
  //  Y
  //--------------------------------------
  // Down-Scaling V : [Y]
  src       = pcYuvSrc->getLumaAddr();
  srcSizeX  = pcYuvSrc->getWidth();
  srcSizeY  = pcYuvSrc->getHeight();
  srcStride = pcYuvSrc->getStride();

  dst       = m_pTemp + m_offset;
  dstSizeX  = pcYuvSrc->getWidth();
  dstSizeY  = pcYuvDst->getHeight();
  dstStride = m_stride;

  if ( m_stFilterV.nbPhase > 1 )  filter_ARC_upScale_phasis(src, srcStride, dst, dstStride, dstSizeX, dstSizeY, &m_stFilterV, 0);
  else                            filter_ARC(src, srcStride, dst, dstStride, dstSizeX, dstSizeY, &m_stFilterV);

  extendBorderLeftRight(dst, dstStride, dstSizeX, dstSizeY, m_marginX);

  // Down-Scaling H : [Y]
  src       = dst;
  srcSizeX  = dstSizeX;
  srcSizeY  = dstSizeY;
  srcStride = dstStride;

  dst       = pcYuvDst->getLumaAddr();
  dstSizeX  = pcYuvDst->getWidth();
  dstSizeY  = pcYuvDst->getHeight();
  dstStride = pcYuvDst->getStride();

  if ( m_stFilterH.nbPhase > 1 )  filter_ARC_upScale_phasis(src, srcStride, dst, dstStride, dstSizeX, dstSizeY, &m_stFilterH, 0);
  else                            filter_ARC(src, srcStride, dst, dstStride, dstSizeX, dstSizeY, &m_stFilterH);


  razTemp();

  //--------------------------------------
  //  U
  //--------------------------------------
  // Down-Scaling V : [U]
  src       = pcYuvSrc->getCrAddr();
  srcSizeX  = pcYuvSrc->getWidth()  >> 1;
  srcSizeY  = pcYuvSrc->getHeight() >> 1;
  srcStride = pcYuvSrc->getCStride();

  dst       = m_pTemp + m_offset;
  dstSizeX  = pcYuvSrc->getWidth()  >> 1;
  dstSizeY  = pcYuvDst->getHeight() >> 1;
  dstStride = m_stride;

  if ( m_stFilterV.nbPhase > 1 )  filter_ARC_upScale_phasis(src, srcStride, dst, dstStride, dstSizeX, dstSizeY, &m_stFilterV, 0);
  else                            filter_ARC(src, srcStride, dst, dstStride, dstSizeX, dstSizeY, &m_stFilterV);

  extendBorderLeftRight(dst, dstStride, dstSizeX, dstSizeY, m_marginX);

  // Down-Scaling H : [U]
  src       = dst;
  srcSizeX  = dstSizeX;
  srcSizeY  = dstSizeY;
  srcStride = dstStride;

  dst       = pcYuvDst->getCrAddr();
  dstSizeX  = pcYuvDst->getWidth()  >> 1;
  dstSizeY  = pcYuvDst->getHeight() >> 1;
  dstStride = pcYuvDst->getCStride();

  if ( m_stFilterV.nbPhase > 1 )  filter_ARC_upScale_phasis(src, srcStride, dst, dstStride, dstSizeX, dstSizeY, &m_stFilterH, 0);
  else                            filter_ARC(src, srcStride, dst, dstStride, dstSizeX, dstSizeY, &m_stFilterH);

  razTemp();

  //--------------------------------------
  //  V
  //--------------------------------------
  // Down-Scaling V : [V]
  src       = pcYuvSrc->getCbAddr();
  srcSizeX  = pcYuvSrc->getWidth()  >> 1;
  srcSizeY  = pcYuvSrc->getHeight() >> 1;
  srcStride = pcYuvSrc->getCStride();

  dst       = m_pTemp + m_offset;
  dstSizeX  = pcYuvSrc->getWidth()  >> 1;
  dstSizeY  = pcYuvDst->getHeight() >> 1;
  dstStride = m_stride;

  if ( m_stFilterV.nbPhase > 1 )  filter_ARC_upScale_phasis(src, srcStride, dst, dstStride, dstSizeX, dstSizeY, &m_stFilterV, 0);
  else                            filter_ARC(src, srcStride, dst, dstStride, dstSizeX, dstSizeY, &m_stFilterV);

  extendBorderLeftRight(dst, dstStride, dstSizeX, dstSizeY, m_marginX);

  // Down-Scaling H : [V]
  src       = dst;
  srcSizeX  = dstSizeX;
  srcSizeY  = dstSizeY;
  srcStride = dstStride;

  dst       = pcYuvDst->getCbAddr();
  dstSizeX  = pcYuvDst->getWidth()  >> 1;
  dstSizeY  = pcYuvDst->getHeight() >> 1;
  dstStride = pcYuvDst->getCStride();

  if ( m_stFilterH.nbPhase > 1 )  filter_ARC_upScale_phasis(src, srcStride, dst, dstStride, dstSizeX, dstSizeY, &m_stFilterH, 0);
  else                            filter_ARC(src, srcStride, dst, dstStride, dstSizeX, dstSizeY, &m_stFilterH);

  razTemp();
}

Void  getPos( TComDataCU *cu, UInt partAddr, TComYuv *curPic, Int yuv, Int &posX, Int &posY)
{
  Pel *curAddr, *startAddr;
  Int stride;

  if ( yuv == Y ) {
    curAddr   = curPic->getLumaAddr( partAddr );
    startAddr = curPic->getLumaAddr();
    stride    = curPic->getStride();
  }
  else if ( yuv == U ) {
    curAddr   = curPic->getCbAddr( partAddr );
    startAddr = curPic->getCbAddr();
    stride    = curPic->getCStride();
  }
  else {
    curAddr   = curPic->getCrAddr( partAddr );
    startAddr = curPic->getCrAddr();
    stride    = curPic->getCStride();
  }

  posY = ( curAddr - startAddr ) / stride;
  posX = curAddr - startAddr - posY*stride;
}

Int ARCfilter::checkLimits(Int posX, Int srcSizeX, Int picSizeX, Int halfFilterSize, Int marginX)
{
  if ( (posX + srcSizeX + halfFilterSize) > (picSizeX+marginX) )  posX -= (posX + srcSizeX + halfFilterSize) - (picSizeX+marginX);
  if ( (halfFilterSize - posX) > marginX )                        posX += halfFilterSize - posX - marginX;

  return( posX );
}

Void  ARCfilter::xPredInterLumaBlk( TComDataCU *cu, TComPicYuv *refPic, UInt partAddr, TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi , Int mode)
{
  Int refStride = refPic->getStride();  
  
  Int xFrac = mv->getHor() & 0x3;
  Int yFrac = mv->getVer() & 0x3;

  assert(xFrac >= 0 && xFrac < 4);
  assert(yFrac >= 0 && yFrac < 4);

  Int tmpSizeX = width<<1,
      tmpSizeY = height<<1;
  setSize(tmpSizeX, tmpSizeY);

  if ( mode == DOWN ) {
    m_downScaleFilter               = g_Luma_downScale2x2_mvInter_filter[xFrac];
    m_downScaleFilter_ifFilterPrec  = g_Luma_downScale2x2_mvInter_ifFilterPrec[xFrac];
    m_downScaleFilter_nTaps         = g_Luma_downScale2x2_mvInter_nbCoeff[xFrac];
  }
  else if ( mode == UP ) {
    assert( g_Luma_upScale2x2_mvInter_nbCoeff[0][xFrac]      == g_Luma_upScale2x2_mvInter_nbCoeff[1][xFrac] );
    assert( g_Luma_upScale2x2_mvInter_ifFilterPrec[0][xFrac] == g_Luma_upScale2x2_mvInter_ifFilterPrec[1][xFrac] );
    for ( int phase=0 ; phase<2 ; phase++ ) {
      m_upScaleFilter[phase]              = g_Luma_upScale2x2_mvInter_filter[phase][xFrac];
      m_upScaleFilter_ifFilterPrec        = g_Luma_upScale2x2_mvInter_ifFilterPrec[phase][xFrac];
      m_upScaleFilter_nTaps               = g_Luma_upScale2x2_mvInter_nbCoeff[phase][xFrac];
    }
  }
  else {
    printf("ARCfilter::xPredInterLumaBlk() => Not supported AR mode.\n");
    exit(0);
  }
  setFilter( mode,  true, false,  false, m_stFilterH);


  if ( mode == DOWN ) {
    m_downScaleFilter               = g_Luma_downScale2x2_mvInter_filter[yFrac];
    m_downScaleFilter_ifFilterPrec  = g_Luma_downScale2x2_mvInter_ifFilterPrec[yFrac];
    m_downScaleFilter_nTaps         = g_Luma_downScale2x2_mvInter_nbCoeff[yFrac];
  }
  else if ( mode == UP ) {
    assert( g_Luma_upScale2x2_mvInter_nbCoeff[0][yFrac]      == g_Luma_upScale2x2_mvInter_nbCoeff[1][yFrac] );      // "nbCoeff" should be the same for all phases
    assert( g_Luma_upScale2x2_mvInter_ifFilterPrec[0][yFrac] == g_Luma_upScale2x2_mvInter_ifFilterPrec[1][yFrac] ); // "ifFilterPrec" should be the same for all phases
    for ( int phase=0 ; phase<2 ; phase++ ) {
      m_upScaleFilter[phase]              = g_Luma_upScale2x2_mvInter_filter[phase][yFrac];
      m_upScaleFilter_ifFilterPrec        = g_Luma_upScale2x2_mvInter_ifFilterPrec[phase][yFrac];
      m_upScaleFilter_nTaps               = g_Luma_upScale2x2_mvInter_nbCoeff[phase][yFrac];
    }
  }
  else {
    printf("ARCfilter::xPredInterLumaBlk() => Not supported AR mode.\n");
    exit(0);
  }
  setFilter( mode, false,  !bi,   true, m_stFilterV);

  int   srcSizeX, srcSizeY, srcStride;
  int   dstSizeX, dstSizeY, dstStride;
  Short *src, *dst;

  razTemp(tmpSizeX, tmpSizeY);

  //--------------------------------------
  //  Y
  //--------------------------------------

  // Get position in Current (dstPic) Picture:
  Int posX, posY;
  posX = cu->getCUPelX() + g_auiRasterToPelX[ g_auiZscanToRaster[partAddr] ];
  posY = cu->getCUPelY() + g_auiRasterToPelY[ g_auiZscanToRaster[partAddr] ];

  posX  += mv->getHor() >> 2;
  posY  += mv->getVer() >> 2;

  Int phasis1_X = abs(posX) % 2;
  Int phasis1_Y = abs(posY) % 2;

  // Get position in Reference (refPic) Picture:
  posX  = ( mode == DOWN )  ? (posX<<1)  : (posX>>1);
  posY  = ( mode == DOWN )  ? (posY<<1)  : (posY>>1);

  Int halfFilterSize  = ( m_stFilterH.N[0] >> 1 ) + 1;
  Int filterSize      = ( m_stFilterH.N[0] );

  // Filtering H : [Y]
  srcSizeX  = ( mode == DOWN )  ? (width<<1)  : (width>>1);
  srcSizeY  = ( mode == DOWN )  ? (height<<1) : (height>>1);
  srcStride = refStride;
  posX      = checkLimits(posX, srcSizeX, refPic->getWidth(),  halfFilterSize, refPic->getLumaMargin() );
  posY      = checkLimits(posY, srcSizeY, refPic->getHeight(), halfFilterSize, refPic->getLumaMargin() );
  src       = refPic->getLumaAddr() + posY*refPic->getStride() + posX;

  dst       = m_pTemp + m_offset;
  dstSizeX  = width;
  dstSizeY  = srcSizeY;
  dstStride = m_stride;

  assert( halfFilterSize < m_marginY );

  if ( m_stFilterH.nbPhase > 1 )  filter_ARC_upScale_phasis(src - halfFilterSize*srcStride, srcStride, dst - halfFilterSize*dstStride, dstStride, dstSizeX, dstSizeY + filterSize, &m_stFilterH, phasis1_X);
  else                            filter_ARC(src - halfFilterSize*srcStride, srcStride, dst - halfFilterSize*dstStride, dstStride, dstSizeX, dstSizeY + filterSize, &m_stFilterH);

  // Filtering V : [Y]
  src       = dst;
  srcSizeX  = dstSizeX;
  srcSizeY  = dstSizeY;
  srcStride = dstStride;

  dst       = dstPic->getLumaAddr( partAddr );
  dstSizeX  = width;
  dstSizeY  = height;
  dstStride = dstPic->getStride();

  if ( m_stFilterV.nbPhase > 1 )  filter_ARC_upScale_phasis(src, srcStride, dst, dstStride, dstSizeX, dstSizeY, &m_stFilterV, phasis1_Y);
  else                            filter_ARC(src, srcStride, dst, dstStride, dstSizeX, dstSizeY, &m_stFilterV);

}

Void  ARCfilter::xPredInterChromaBlk( TComDataCU *cu, TComPicYuv *refPic, UInt partAddr, TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi , Int mode)
{
  Int refStride = refPic->getCStride();  
  
  Int     xFrac  = mv->getHor() & 0x7;
  Int     yFrac  = mv->getVer() & 0x7;
  assert(xFrac >= 0 && xFrac < 8);
  assert(yFrac >= 0 && yFrac < 8);

  UInt    cxWidth  = width  >> 1;
  UInt    cxHeight = height >> 1;

  Int tmpSizeX = cxWidth<<1,
      tmpSizeY = cxHeight<<1;
  setSize(cxWidth<<1, cxHeight<<1);

  // Horizontal Filter settings :
  if ( mode == DOWN ) {
    m_downScaleFilter               = g_Chroma_downScale2x2_mvInter_filter[xFrac];
    m_downScaleFilter_ifFilterPrec  = g_Chroma_downScale2x2_mvInter_ifFilterPrec[xFrac];
    m_downScaleFilter_nTaps         = g_Chroma_downScale2x2_mvInter_nbCoeff[xFrac];
  }
  else if ( mode == UP ) {
    assert( g_Chroma_upScale2x2_mvInter_nbCoeff[0][xFrac]      == g_Chroma_upScale2x2_mvInter_nbCoeff[1][xFrac] );
    assert( g_Chroma_upScale2x2_mvInter_ifFilterPrec[0][xFrac] == g_Chroma_upScale2x2_mvInter_ifFilterPrec[1][xFrac] );
    for ( int phase=0 ; phase<2 ; phase++ ) {
      m_upScaleFilter[phase]              = g_Chroma_upScale2x2_mvInter_filter[phase][xFrac];
      m_upScaleFilter_ifFilterPrec        = g_Chroma_upScale2x2_mvInter_ifFilterPrec[phase][xFrac];
      m_upScaleFilter_nTaps               = g_Chroma_upScale2x2_mvInter_nbCoeff[phase][xFrac];
    }
  }
  else {
    printf("ARCfilter::xPredInterLumaBlk() => Not supported AR mode.\n");
    exit(0);
  }
  setFilter( mode,  true, false,  false, m_stFilterH);

  // Vertical Filter settings :
  if ( mode == DOWN ) {
    m_downScaleFilter               = g_Chroma_downScale2x2_mvInter_filter[yFrac];
    m_downScaleFilter_ifFilterPrec  = g_Chroma_downScale2x2_mvInter_ifFilterPrec[yFrac];
    m_downScaleFilter_nTaps         = g_Chroma_downScale2x2_mvInter_nbCoeff[yFrac];
  }
  else if ( mode == UP ) {
    assert( g_Chroma_upScale2x2_mvInter_nbCoeff[0][yFrac]      == g_Chroma_upScale2x2_mvInter_nbCoeff[1][yFrac] );
    assert( g_Chroma_upScale2x2_mvInter_ifFilterPrec[0][yFrac] == g_Chroma_upScale2x2_mvInter_ifFilterPrec[1][yFrac] );
    for ( int phase=0 ; phase<2 ; phase++ ) {
      m_upScaleFilter[phase]              = g_Chroma_upScale2x2_mvInter_filter[phase][yFrac];
      m_upScaleFilter_ifFilterPrec        = g_Chroma_upScale2x2_mvInter_ifFilterPrec[phase][yFrac];
      m_upScaleFilter_nTaps               = g_Chroma_upScale2x2_mvInter_nbCoeff[phase][yFrac];
    }
  }
  else {
    printf("ARCfilter::xPredInterLumaBlk() => Not supported AR mode.\n");
    exit(0);
  }
  setFilter( mode, false,  !bi,   true, m_stFilterV);


  int   srcSizeX, srcSizeY, srcStride;
  int   dstSizeX, dstSizeY, dstStride;
  Short *src, *dst;

  razTemp(tmpSizeX, tmpSizeY);

  //--------------------------------------
  //  U
  //--------------------------------------
  // Get position in Current (dstPic) Picture:
  Int posX, posY;
  posX = cu->getCUPelX() + g_auiRasterToPelX[ g_auiZscanToRaster[partAddr] ];
  posY = cu->getCUPelY() + g_auiRasterToPelY[ g_auiZscanToRaster[partAddr] ];
  posX = posX >> 1;
  posY = posY >> 1;

  posX  += mv->getHor() >> 3;
  posY  += mv->getVer() >> 3;

  Int phasis1_X = abs(posX) % 2;
  Int phasis1_Y = abs(posY) % 2;

  // Get position in Reference (refPic) Picture:
  posX  = ( mode == DOWN )  ? (posX<<1)  : (posX>>1);
  posY  = ( mode == DOWN )  ? (posY<<1)  : (posY>>1);

  Int halfFilterSize  = ( m_stFilterH.N[0] >> 1 ) + 1;
  Int filterSize      = ( m_stFilterH.N[0] );

  // Filtering H : [U]
  srcSizeX  = ( mode == DOWN )  ? (cxWidth<<1)  : (cxWidth>>1);
  srcSizeY  = ( mode == DOWN )  ? (cxHeight<<1) : (cxHeight>>1);
  srcStride = refStride;
  posX      = checkLimits(posX, srcSizeX, refPic->getWidth()>>1,  halfFilterSize, refPic->getChromaMargin() );
  posY      = checkLimits(posY, srcSizeY, refPic->getHeight()>>1, halfFilterSize, refPic->getChromaMargin() );
  src       = refPic->getCbAddr() + posY*refPic->getCStride() + posX;

  dst       = m_pTemp + m_offset;
  dstSizeX  = cxWidth;
  dstSizeY  = srcSizeY;
  dstStride = m_stride;

  assert( halfFilterSize < m_marginY );

  if ( m_stFilterH.nbPhase > 1 )  filter_ARC_upScale_phasis(src - halfFilterSize*srcStride, srcStride, dst - halfFilterSize*dstStride, dstStride, dstSizeX, dstSizeY + filterSize, &m_stFilterH, phasis1_X);
  else                            filter_ARC(src - halfFilterSize*srcStride, srcStride, dst - halfFilterSize*dstStride, dstStride, dstSizeX, dstSizeY + filterSize, &m_stFilterH);

  // Filtering V : [U]
  src       = dst;
  srcSizeX  = dstSizeX;
  srcSizeY  = dstSizeY;
  srcStride = dstStride;

  dst       = dstPic->getCbAddr( partAddr );
  dstSizeX  = cxWidth;
  dstSizeY  = cxHeight;
  dstStride = dstPic->getCStride();

  if ( m_stFilterV.nbPhase > 1 )  filter_ARC_upScale_phasis(src, srcStride, dst, dstStride, dstSizeX, dstSizeY, &m_stFilterV, phasis1_Y);
  else                            filter_ARC(src, srcStride, dst, dstStride, dstSizeX, dstSizeY, &m_stFilterV);

  razTemp(tmpSizeX, tmpSizeY);

  //--------------------------------------
  //  V
  //--------------------------------------
  // Filtering H : [V]
  srcSizeX  = ( mode == DOWN )  ? (cxWidth<<1)  : (cxWidth>>1);
  srcSizeY  = ( mode == DOWN )  ? (cxHeight<<1) : (cxHeight>>1);
  srcStride = refStride;
  src       = refPic->getCrAddr() + posY*refPic->getCStride() + posX;

  dst       = m_pTemp + m_offset;
  dstSizeX  = cxWidth;
  dstSizeY  = srcSizeY;
  dstStride = m_stride;

  if ( m_stFilterH.nbPhase > 1 )  filter_ARC_upScale_phasis(src - halfFilterSize*srcStride, srcStride, dst - halfFilterSize*dstStride, dstStride, dstSizeX, dstSizeY + filterSize, &m_stFilterH, phasis1_X);
  else                            filter_ARC(src - halfFilterSize*srcStride, srcStride, dst - halfFilterSize*dstStride, dstStride, dstSizeX, dstSizeY + filterSize, &m_stFilterH);

  // Filtering V : [V]
  src       = dst;
  srcSizeX  = dstSizeX;
  srcSizeY  = dstSizeY;
  srcStride = dstStride;

  dst       = dstPic->getCrAddr( partAddr );
  dstSizeX  = cxWidth;
  dstSizeY  = cxHeight;
  dstStride = dstPic->getCStride();

  if ( m_stFilterV.nbPhase > 1 )  filter_ARC_upScale_phasis(src, srcStride, dst, dstStride, dstSizeX, dstSizeY, &m_stFilterV, phasis1_Y);
  else                            filter_ARC(src, srcStride, dst, dstStride, dstSizeX, dstSizeY, &m_stFilterV);

}
