#ifndef __ARCFILTERS__
#define __ARCFILTERS__

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "TLibCommon/CommonDef.h"
#include "TLibCommon/TComPicYuv.h"
#include "TLibCommon/TComInterpolationFilter.h"
#include "TLibCommon/TComPrediction.h"

#define DOWN  (0)
#define UP    (1)
#define SAME  (2)

#define NB_PHASE  (2)
typedef struct {
  Short *coeff[NB_PHASE]; // filter coefficients (phase 0 and 1)
  Int   N[NB_PHASE],      // nb tap filters (phase 0 and 1)
        nbPhase;          // nbPhase <= NB_PHASE
  Bool  isVertical,
        isFirst, 
        isLast;
  Int cs, rs;       // column and row steps in src
  Int ifFilterPrec;
} StructFilterARC;

//======================================================
// Up/Down-scaling Filters
//======================================================
#define NTAPS_DOWN_SCALE    (12)
//downScale2x2
extern  Short m_downScaleFilter[NTAPS_DOWN_SCALE];
extern  Int   m_downScaleFilter_ifFilterPrec;

#define NTAPS_UP_SCALE    (4)
//upScale2x2
extern  Short m_upScaleFilter[2][NTAPS_UP_SCALE];
extern  Int m_upScaleFilter_ifFilterPrec;

// ====================================================================================================================
// Class definition
// ====================================================================================================================

/// Decoder configuration class
class ARCfilter
{
  Pel*    m_pTemp;
  Int     m_sizeX, m_sizeY, m_marginX, m_marginY,
          m_stride, m_offset;

  Short*  m_downScaleFilter;
  Int     m_downScaleFilter_ifFilterPrec;
  Int     m_downScaleFilter_nTaps;

  Short*  m_upScaleFilter[2];
  Int     m_upScaleFilter_ifFilterPrec; // should same for all phasis
  Int     m_upScaleFilter_nTaps;        // should same for all phasis

  StructFilterARC m_stFilterH, m_stFilterV;

protected:
  Void  setSize(Int width, Int height);
  Void  razTemp();
  Void  razTemp(Int Sx, Int Sy);

  Int   checkLimits(Int posX, Int srcSizeX, Int picSizeX, Int halfFilterSize, Int marginX);

  Void  extendBorderUpDown(Pel *src, Int stride, Int width, Int height, Int marginY);
  Void  extendBorderLeftRight(Pel *src, Int stride, Int width, Int height, Int marginX);

  Void  setFilter( Int mode, Bool isFirst, Bool isLast, Bool isVertical, StructFilterARC &stFilter);

  Void  filter( TComPicYuv* pcYuvSrc, TComPicYuv* pcYuvDst , Int mode, Bool isRef);
  
  Void  filter_ARC(Short const *src, Int srcStride, Short *dst, Int dstStride, Int width, Int height, StructFilterARC *stFilter);
  Void  filter_ARC_upScale_phasis(Short const *src, Int srcStride, Short *dst, Int dstStride, Int width, Int height, StructFilterARC *stFilter, Int phasis1st);


public:
  ARCfilter(Int width, Int height);
  ~ARCfilter();

  Void  freeTmpBuffer();

  Void  scalePic( TComPicYuv* pcYuvSrc, TComPicYuv* pcYuvDst , Int mode, Bool isRef);

  Int   getMode( Int iCurSizeIdx, Int iRefSizeIdx );

  Void  xPredInterLumaBlk  ( TComDataCU *cu, TComPicYuv *refPic, UInt partAddr, TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi, Int mode );
  Void  xPredInterChromaBlk( TComDataCU *cu, TComPicYuv *refPic, UInt partAddr, TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi , Int mode);
};

extern  ARCfilter *g_arcFilter;

//! \}

#endif



